import os
import numpy as np
import pandas as pd
import scanpy as sc
import networkx as nx
from anndata import AnnData
from pandas.api.types import is_categorical_dtype
from typing import Union, Optional, Literal

from ._preprocess import Preprocessing
from ._tree_alignment import Tree_Alignment
from ._dynamic_time_warping import DPT, DynamicTimeWarping
from .._util import CapitalData


def preprocessing(
    adata: AnnData,
    Min_Genes: int = 200,
    Min_Cells: int = 3,
    Min_Mean: float = 0.0125,
    Max_Mean: float = 3,
    Min_Disp: float = 0.5,
    N_pcs: int = 50,
    n_Top_genes: int = 2000,
    K: int = 10,
    magic_imputation: bool = False,
):
    """\
    The recipe for preprocessing raw count data.

    In adata.raw, all genes are stored, so that those genes can be used in later calculation in CAPITAL.

    The recipe runs the following steps:
    ::

        import scanpy as sc
        sc.pp.filter_cells(adata, min_genes=Min_Genes)
        sc.pp.filter_genes(adata, min_cells=Min_Cells)
        sc.pp.normalize_total(adata, exclude_highly_expressed=True)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, min_mean=Min_Mean, max_mean=Max_Mean, min_disp=Min_Disp, n_top_genes=n_Top_genes)
        adata.raw = adata
        adata = adata[:,adata.var['highly_variable']]
        sc.tl.pca(adata, n_comps=N_pcs)
        sc.pp.neighbors(adata, n_neighbors=K, n_pcs=N_pcs)
        sc.tl.diffmap(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata)
        sc.tl.paga(adata, groups='leiden')

    Parameters
    ----------
    adata : AnnData
        The annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    Min_Genes : int
        The number of genes to filter in scanpy.pp.filter_cells(),
        by default 200.
    Min_Cells : int
        The number of cells to filter in scanpy.pp.filter_genes(),
        by default 3.
    Min_Mean : float
        The minimum mean that is filtered to calculate highly variable genes.
        Look scanpy.pp.highly_variable_genes(),
        by default 0.0125.
    Max_Mean : int
        The maxmum mean that is filtered to calculate highly variable genes.
        Look scanpy.pp.highly_variable_genes(),
        by default 3.
    Min_Disp : float
        The minimum dispersion that is filtered to calculate highly variable genes.
        Look scanpy.pp.highly_variable_genes(),
        by default 0.5.
    N_pcs : int
        The number of principal components used,
        by default 50.
    n_Top_genes : int
        The number of highly variable genes,
        by default 2000.
    K : int
        The size of a local neighborhood used for manifold approximation,
        by default 10.
    magic_imputation : bool
        If `True`, MAGIC imputation is done,
        by default `False`.
    """

    if not isinstance(adata, AnnData):
        raise ValueError("preprocessing() expects AnnData argument")

    pp = Preprocessing()
    pp.preprocessing_rawdata(
        adata,
        Min_Genes=Min_Genes,
        Min_Cells=Min_Cells,
        Min_Mean=Min_Mean,
        Max_Mean=Max_Mean,
        Min_Disp=Min_Disp,
        N_pcs=N_pcs,
        n_Top_genes=n_Top_genes,
        K=K,
        magic_imputation=magic_imputation,
        copy=False
    )


def trajectory_tree(
    adata: AnnData,
    root_node: Optional[str] = None,
    method: Literal["euclid", "gauss", "paga"] = "euclid",
    dimension: Literal["pca", "diffmap"] ="pca",
    tree: Optional[nx.DiGraph] = None,
    groupby: str = "leiden"
):
    """\
    Calculate a trajectory tree.

    Specify AnnData and the root name in the trajectory.

    You can pass your metadata in adata.obs
    by passing a key in adata.obs to `groupby`.

    You can also define your own tree by drawing it as nx.DiGraph
    and pass it to `tree`.

    Parameters
    ----------
    adata : AnnData
        The annotated data matrix.
    root_node : Optional[str]
        A given cluster is used as the root for the tree.
        If `None`, it calculates a most likely root of the tree,
        but this will not be reliable for the moment.
        Please set the root cluster, by default `None`.
    method : Literal["euclid", "gauss", "paga"]
        Method for calculating the tree, by default "euclid."
    dimension : Literal["pca", "diffmap"]
        The data for calculating the trajectory tree, by default "pca."
    tree : Optional[nx.DiGraph]
        If `None`, it calculates the trajectory tree. If nx.DiGraph is given,
        the tree is constructed from the nx.DiGraph, by default `None`.
    groupby : str
        Key for categorical in `adata.obs`. You can pass your
        metadata of clusters, by default "leiden."
    """
    if method not in ["euclid", "paga", "gauss"]:
        raise ValueError("Argument 'method' must be 'euclid','paga' or 'gauss'.")
    if dimension not in ['pca', 'diffmap']:
        raise ValueError("Argument 'dimension' must be 'pca' or 'diffmap'.")

    if not isinstance(adata, AnnData):
        raise ValueError("trajectory_tree() expects AnnData argument")
    else:
        if groupby not in adata.obs.keys():
            raise ValueError("Did not find adata.obs[{}]".format(groupby))
        if "X_{}".format(dimension) not in list(adata.obsm):
            raise ValueError(
                "Did not find 'X_{}' in adata.obsm. Run scanpy.tl to calculate first.".format(dimension))

    # when adata.obs[groupby] is metadata and not categorized, it cause errors.
    # maybe there are better ways to set adata.obs, but one code below does jobs enough.
    if not is_categorical_dtype(adata.obs[groupby]):
        adata.obs[groupby] = adata.obs[groupby].astype('category')

    pp = Preprocessing()
    pp.trajectory_tree(
        adata,
        root_node=root_node,
        tree=tree,
        method=method,
        dimension=dimension,
        groupby=groupby
    )


def _tree_align_dpt(
    adata1: AnnData,
    adata2: AnnData,
    num_genes1: int = 2000,
    num_genes2: int = 2000,
    cost: float = 1.0,
    use_raw: bool = True
):
    """\
    Just a combination of tree_alignment() and dpt().

    Parameters
    ----------
    adata1 : AnnData
        [description]
    adata2 : AnnData
        [description]
    num_genes1 : int, optional
        [description], by default 2000.
    num_genes2 : int, optional
        [description], by default 2000.
    cost : float, optional
        [description], by default 1.0.
    use_raw : bool, optional
        [description], by default `True`.

    Returns
    -------
    capital_data: [CapitalData]
        [description]

    """
    if not isinstance(adata1, AnnData):
        raise ValueError("tree_align() expect AnnData argument")
    if not isinstance(adata2, AnnData):
        raise ValueError("tree_align() expect AnnData argument")

    if 'capital' not in adata1.uns:
        raise ValueError(
            "'capital' is not found in adata1.uns. Run cp.trajectory_tree() first.")

    if 'capital' not in adata2.uns:
        raise ValueError(
            "'capital' is not found in adata2.uns. Run cp.trajectory_tree() first.")

    if use_raw:
        if adata1.raw == None:
            adata1.raw = adata1
        if adata2.raw == None:
            adata2.raw = adata2
    else:
        adata1.raw = adata1
        adata2.raw = adata2

    print("Calculating tree alignment")
    tree_align = Tree_Alignment()
    aligned_data = tree_align.tree_alignment(
        adata1,
        adata2,
        cost=cost,
        N_1=num_genes1,
        N_2=num_genes2
    )
    print("Calculating pseudotime for each alignment")
    dpt = DPT()
    dpt.dpt_for_alignments(
        aligned_data,
        no_prune=False
    )

    print("Calculation finished.")

    return aligned_data


def tree_alignment(
    adata1: AnnData,
    adata2: AnnData,
    num_genes1: int = 2000,
    num_genes2: int = 2000,
    cost: float = 1.0,
    use_raw: bool = True
):
    """\
    Calculate an alignment of the two trajectory trees.

    Use a dynamic programming algorithm in `[P Bille05] <https://www.sciencedirect.com/science/article/pii/S0304397505000174>`_ .
    Use `capital.pl.tree_alignment` to obtain the alignment tree.

    Parameters
    ----------
    adata1 : AnnData
        The annotated data matrix.
    adata2 : AnnData
        The annotated data matrix.
    num_genes1 : int
        Number of highly variable genes in adata1 to calculate intersection of the genes.
        If not `None`, it recalculates highly variable genes using genes in adata.raw,
        if `None`, it uses the genes in adata.obs,
        by default 2000.
    num_genes2 : int
        Same as num_genes1, but for adata2, by default 2000.
    cost : float
        Gap cost for the tree alignment calculation, by default 1.0.
    use_raw : bool
        If `True`, it uses adata.raw to calculate highly variable genes.
        If `False` and adata.raw was `None`, it resets adata.raw as adata, by default `True`.

    Returns
    -------
    capital_data: [CapitalData]
        The data matrices containing the results of CAPTIAL.

    """
    if not isinstance(adata1, AnnData):
        raise ValueError("tree_align() expect AnnData argument")
    if not isinstance(adata2, AnnData):
        raise ValueError("tree_align() expect AnnData argument")

    if 'capital' not in adata1.uns:
        raise ValueError("'capital' is not found in adata1.uns. Run cp.trajectory_tree() first.")

    if 'capital' not in adata2.uns:
        raise ValueError("'capital' is not found in adata2.uns. Run cp.trajectory_tree() first.")

    if use_raw:
        if adata1.raw == None:
            adata1.raw = adata1
        if adata2.raw == None:
            adata2.raw = adata2
    else:
        adata1.raw = adata1
        adata2.raw = adata2

    print("Calculating tree alignment")
    tree_align = Tree_Alignment()
    aligned_data = tree_align.tree_alignment(
        adata1,
        adata2,
        cost=cost,
        N_1=num_genes1,
        N_2=num_genes2
    )
    print("Calculation finished.")

    return aligned_data

def dpt(
    aligned_data: CapitalData,
    alignment: Union[str, list, None] = None,
    no_prune: bool = False
):
    """\
    Calculate pseudotime for the alignments.

    Use `scanpy.tl.dpt() <https://scanpy.readthedocs.io/en/stable/api/scanpy.tl.dpt.html#scanpy.tl.dpt>`_.

    If there are many alignments, you are recommended to specify which alignment is used to calculate pseudotime to save time.
    Use `cdata.alignmentlist` and `capital.pl.tree_alignment` to take a look at alignment data.

    Parameters
    ----------
    aligned_data : CapitalData
        The data matrices containing the results of CAPTIAL.
    alignment : Union[str, list, None]
        Specify which alignment is used to calculate pseudotime.
        If `None`, it calculates pseudotime for all alignments, by default `None`.
    no_prune : bool
        Specify global alignment or local alignment.
        If `True`, it calculates pseudotime using all the cells in the clusters in the alignment.
        If `False`, it uses all the cells except ones that have no corresponding cluster in the root or the leaf of the alignment,
        by default `False`.
    """

    if not isinstance(aligned_data, CapitalData):
        raise ValueError(
                        "dpt() expect CapitalData argument."
                        "Run tree_alignment() first and pass the return as the argument as below.\n"
                        "cdata = cp.tree_alignment(adata1, adata2)\n"
                        "cp.dpt(cdata)\n"
                        )
    dpt = DPT()
    dpt.dpt_for_alignments(
        aligned_data,
        alignment=alignment,
        no_prune=no_prune
    )


def dtw(
    aligned_data: CapitalData,
    gene: Union[str, list, np.ndarray],
    alignment: Union[str, list, None] = None,
    multi_genes=False,
):
    """\
    Calculate dynamic time warping for genes.

    This requires that `dpt()` has been run for the specified alignments.

    Parameters
    ----------
    aligned_data : CapitalData
        The data matrices containing the results of CAPTIAL.
    gene : Union[str, list, np.ndarray]
        Genes for calculating dynamic time warping.
    alignment : Union[str, list, None]
        Specify which alignment is used to calculate dynamic time warping.
        If `None`, it calculates dynamic time warping for all the alignments, by default `None`.
    multi_genes : bool
        If True, compute dynamic time warping using all genes in argument gene and 
        return matched cells in cdata.alignmnetdict['alignmentid']['multi_genes'].
        If False, compute dynamic time warping for each gene. By defalut False.  

    """

    if not isinstance(aligned_data, CapitalData):
        raise ValueError(
                        "dtw() expect CapitalData argument." \
                        "Run tree_align() first and pass the return as the argument as below.\n"\
                        "cdata = cp.tree_alignment(adata1, adata2)\n"\
                        "cp.dpt(cdata)\n"\
                        "cp.dtw(cdata, genelist)"
                        )
    dtw = DynamicTimeWarping()
    dtw.dtw_for_alignments(
        aligned_data,
        gene,
        alignment=alignment,
        multi_genes=multi_genes
    )


def genes_similarity_score(
    aligned_data: CapitalData,
    gene: Union[str, list, np.ndarray, None] = None,
    alignment: Union[str, list, None] = None,
    min_disp: float = 2.0,
):
    """\
    Calculate similarity scores using dynamic time warping `[H.Sakoe78] <https://ieeexplore.ieee.org/document/1163055>`_.

    Use `tslearn.metrics.dtw <https://tslearn.readthedocs.io/en/stable/gen_modules/metrics/tslearn.metrics.dtw.html#r34771c3a90b5-1>`_ \
    for each gene and calculate the similarity of gene expression dynamics.

    Use `cdata.similarity_score` to access the results. Higher on the list, more similar the gene expression dynamics are.

    Parameters
    ----------
    aligned_data : CapitalData
        The data matrices containing the results of CAPTIAL.
    gene : Union[str, list, np.ndarray]
        Genes for calculating similarity, by default `None`.
    alignment : Union[str, list, None]
        Specify alignments to calculate similarity. If `None`, it computes for all the alignments,
        by default `None`.
    min_disp : float
        If `gene` is `None`, the union of genes that have bigger dispersion than min_disp
        in the clusters of the alignment is used, by default 2.0.
    """
    if not isinstance(aligned_data, CapitalData):
        raise ValueError(
            "genes_similarity_score() expect CapitalData argument.")

    dtw = DynamicTimeWarping()
    dtw.get_genes_similarity_score(
        aligned_data,
        gene=gene,
        alignment=alignment,
        min_disp=min_disp,
    )


def read_capital_data(
    dirname: str,
    adata1_name: Optional[str] = None,
    adata2_name: Optional[str] = None
):
    """\
    Reading CAPITAL data.

    Three files (capital_data.npz, adata1.h5ad, adata2.hda5) must be in the same directory.

    Parameters
    ----------
    dirname : str
        Path to the CAPITAL data directory.
    adata1_name : Optional[str]
        Specify adata1 file name if the file name is changed since created, by default `None`.
    adata2_name : Optional[str]
        Specify adata2 file name if the file name is changed since created, by default `None`.
    """
    capital_data = np.load(os.path.join(dirname, "capital_data.npz"), allow_pickle=True)

    filedata = capital_data['filenamedata']
    adata1_tmp_name = filedata[0][0]
    adata2_tmp_name = filedata[1][0]

    adata1_filename = adata1_name if adata1_name is not None else adata1_tmp_name
    adata2_filename = adata2_name if adata2_name is not None else adata2_tmp_name

    adata1 = sc.read(os.path.join(dirname, adata1_filename))
    adata2 = sc.read(os.path.join(dirname, adata2_filename))

    alignedtree_adjacencymatrix = capital_data["alignedtree"]
    alignedtree_node = capital_data["alignedtree_node"]
    df_aligned_tree = pd.DataFrame(alignedtree_adjacencymatrix, index=alignedtree_node, columns=alignedtree_node)
    aligned_tree = nx.convert_matrix.from_pandas_adjacency(df_aligned_tree, create_using=nx.DiGraph)
    genes_for_tree_align = capital_data["genes_for_tree_align"]
    alignmentcost = capital_data["alignmentcost"]
    alignmentdict = capital_data["alignmentdict"][()]
    similarity_score = capital_data["similarity_score"][()]

    alignd_data = CapitalData(
        adata1,
        adata2,
        aligned_tree,
        alignmentcost,
        genes_for_tree_align,
        alignmentdict
        )

    alignd_data.similarity_score = similarity_score

    return alignd_data




from . import tl
from . import pl
from . import dataset

# might be supressing too many future warnings.
# when in tests, delete below two lines to check future warnings.
import warnings
warnings.simplefilter('ignore', FutureWarning)

__copyright__ = 'Copyright (C) 2021 Reiichi Sugihara, Yuki Kato, Tomoya Mori, Yukio Kawahara'
__version__ = '1.0.12'
__license__ = 'BSD-3-Clause'
__author__ = 'Reiichi Sugihara, Yuki Kato, Tomoya Mori, Yukio Kawahara'
__author_email__ = 'reiichi.sugihara@gmail.com'
__url__ = 'https://github.com/ykat0/capital'



import os
import scanpy as sc


def setty19(
    fpath: str = None
):
    """
    A preprocessed dataset of `Setty2019 <https://www.nature.com/articles/s41587-019-0068-4>`_  \
    used in our work.
    This downloads the dataset from `here <http://www.med.osaka-u.ac.jp/pub/rna/ykato/project/capital/>`__.

    Parameters
    ----------
    fpath : str
        If a path is specified as `str`, the dataset is saved in the path.
        If `None`, the dataset is saved in `../data/capital_dataset`, by default `None`.

    Returns
    -------
    An :class:`~anndata.AnnData` object
    """

    URL = "http://www.med.osaka-u.ac.jp/pub/rna/ykato/project/capital/data/setty2019_magic_n2000_k40_r4.h5ad"

    if fpath:
        download_filepath = fpath
    else:
        download_filepath = "../data/capital_dataset/setty19_capital.h5ad"

    if os.path.exists(download_filepath):
        print(f"The dataset already exist in {download_filepath}.")
        print(f"Reading datasets from {download_filepath}.")
        adata = sc.read(download_filepath)

    else:
        os.makedirs(os.path.dirname(download_filepath), exist_ok=True)
        print("Downloading the dataset.")
        adata = sc.read(download_filepath, backup_url=URL)
        print("Download completed. The dataset is saved in {}".format(download_filepath))

    return adata


def paul15(
    fpath: str = None
):
    """
    A preprocessed dataset of `Paul15 <https://www.cell.com/cell/fulltext/S0092-8674(15)01493-2>`_  \
    used in our work.
    This downloads the dataset from `here <https://github.com/ykat0/capital/tree/master/data>`__.


    Parameters
    ----------
    fpath : str
        If a path is specified as `str`, the dataset is saved in the path.
        If `None`, the dataset is saved in `../data/capital_dataset`, by default `None`.

    Returns
    -------
    An :class:`~anndata.AnnData` object
    """

    URL = "https://github.com/ykat0/capital/raw/master/data/paul2015_magic_n500_k10_r7.h5ad"

    if fpath:
        download_filepath = fpath
    else:
        download_filepath = "../data/capital_dataset/paul15_capital.h5ad"

    if os.path.exists(download_filepath):
        print(f"The dataset already exist in {download_filepath}.")
        print(f"Reading datasets from {download_filepath}.")
        adata = sc.read(download_filepath)

    else:
        os.makedirs(os.path.dirname(download_filepath), exist_ok=True)
        print("Downloading the dataset.")
        adata = sc.read(download_filepath, backup_url=URL)
        print("Download completed. The dataset is saved in {}".format(download_filepath))

    return adata


def velten17(
    fpath: str = None
):
    """
    A preprocessed dataset of `Velten2017 <https://www.nature.com/articles/ncb3493>`_  \
    used in our work.
    This downloads the dataset from `here <https://github.com/ykat0/capital/tree/master/data>`__.

    Parameters
    ----------
    fpath : str
        If a path is specified as `str`, the dataset is saved in the path.
        If `None`, the dataset is saved in `../data/capital_dataset`, by default `None`.

    Returns
    -------
    An :class:`~anndata.AnnData` object
    """

    URL = "https://github.com/ykat0/capital/raw/master/data/velten2017_por_magic_n250_k6_r0.h5ad"

    if fpath:
        download_filepath = fpath
    else:
        download_filepath = "../data/capital_dataset/velten17_capital.h5ad"

    if os.path.exists(download_filepath):
        print(f"The dataset already exist in {download_filepath}.")
        print(f"Reading datasets from {download_filepath}.")
        adata = sc.read(download_filepath)

    else:
        os.makedirs(os.path.dirname(download_filepath), exist_ok=True)
        print("Downloading the dataset.")
        adata = sc.read(download_filepath, backup_url=URL)
        print("Download completed. The dataset is saved in {}".format(download_filepath))

    return adata



import os
import networkx as nx
import numpy as np
from anndata import AnnData
import dataclasses
from typing import Final

@dataclasses.dataclass
class CapitalData:
    adata1: Final[AnnData] = None
    adata2: Final[AnnData] = None
    alignedtree: Final[nx.Graph] = None
    alignmentcost: Final[float] = None
    genes_for_tree_align: Final[list] = None
    alignmentdict: dict = dataclasses.field(
        default_factory=dict
    )
    alignmentlist: list = dataclasses.field(
        default_factory=list,
        init=False
    )
    similarity_score: dict = dataclasses.field(
        default_factory=dict,
        init=False
    )

    def __post_init__(self):
        dic = self.alignmentdict
        self.alignmentlist = [
            (i, dic[i]["data1"], dic[i]["data2"])
            for i in dic.keys()
            ]

    def write(
        self,
        dirname=None,
        adata1_name=None,
        adata2_name=None
    ):
        """\
        Write CapitalData.

        Save data in a file. In the file, capital_data.npz, adata1.h5ad and adata2.h5ad are saved.

        Parameters
        ----------
        dirname : str,
            Path to the directory, if `None`, saves in `../results/cdata`, by default None
        adata1_name : str
            File name of adata1, by default None
        adata2_name : str
            File name of adata2, by default None
        """
        if dirname is None:
            dirname = "../results/cdata"
        os.makedirs(dirname, exist_ok=True)

        adata1_filename = adata1_name if adata1_name is not None else "adata1.h5ad"
        adata2_filename = adata2_name if adata2_name is not None else "adata2.h5ad"

        filenamedata = np.array(
                    [[str(adata1_filename)],
                        [str(adata2_filename)],
                        ["capital_data.npz"]],
                    dtype=object
                )
        self.adata1.write(os.path.join(dirname, adata1_filename))
        self.adata2.write(os.path.join(dirname, adata2_filename))

        np.savez_compressed(os.path.join(dirname, "capital_data.npz"),
                            filenamedata=filenamedata,
                            genes_for_tree_align=self.genes_for_tree_align,
                            alignmentcost=self.alignmentcost,
                            alignmentdict=self.alignmentdict,
                            similarity_score=self.similarity_score,
                            alignedtree=nx.convert_matrix.to_numpy_array(self.alignedtree),
                            alignedtree_node=np.array(self.alignedtree.nodes())
                            )



import os
import matplotlib.pyplot as plt
import itertools
import networkx as nx
import numpy as np
import pandas as pd
from matplotlib import rcParams
from matplotlib.lines import Line2D
from networkx.drawing.nx_pydot import graphviz_layout
from anndata import AnnData
from typing import Union, Optional
from .._util import CapitalData


def dtw(
    aligned_data: CapitalData,
    gene: Union[str, list],
    alignment: Union[str, list, None] = None,
    data1_name: Optional[str] = "data1",
    data2_name: Optional[str] = "data2",
    ncols: int = 2,
    widthspace: float = 0.10,
    heightspace: float = 0.30,
    fontsize: float = 12,
    legend_fontsize: float = 12,
    ticksize: float = 12,
    linecolor: str = 'grey',
    dpi: int = 600,
    show: bool = True,
    save: Union[str, bool] = False,
):
    """\
    Plot the results of dynamic time warping.
    When plotting the result of cp.tl.dtw(multi_genes=True), pass "multi_genes" to argument `gene`.

    Parameters
    ----------
    aligned_data : CapitalData
        The data matrices containing the results of CAPTIAL.
    gene : Union[str, list]
        Keys for annotations of genes.
    alignment : Union[str, list, None], optional
        Keys for alignments to be plotted. If `None`, all alignments will be plotted, by default `None`.
    data1_name : Optional[str], optional
        Text of data1's legend, by default "data1".
    data2_name : Optional[str], optional
        Text of data2's legend, by default "data2".
    ncols : int
        Number of panels per row, by default 2.
    widthspace : float
        Width of space in the panels, by default 0.10.
    heightspace : float
        Height of space in the panels, by default 0.30.
    fontsize : float
        Font size of the title, by default 12.
    legend_fontsize : float, optional
        Font size of the legend, by default 12.
    ticksize : float, optional
        Tick size of the x-axis, by default 12.
    linecolor: str
        Color of lines that connect the cells between data1 and data2.
    show : bool,
        If `True`, show the figure. If `False`, return the figure, by default `True`.
    save : Union[str, bool]
        If `True` or `str`, save the figure. If a path is specified as `str`, the figure is saved in the path, by default `False`.
    """

    if not isinstance(aligned_data, CapitalData):
        ValueError("draw_dtw() expects an CapitalData argument.")

    plot_list = __set_plot_list(aligned_data, alignment, gene)

    data1 = aligned_data.adata1
    data2 = aligned_data.adata2
    groupby1 = data1.uns["capital"]["tree"]["annotation"]
    groupby2 = data2.uns["capital"]["tree"]["annotation"]
    groupby_colors1 = groupby1 + "_colors"
    groupby_colors2 = groupby2 + "_colors"

    if groupby_colors1 not in data1.uns:
        from scanpy.plotting._utils import _set_default_colors_for_categorical_obs
        _set_default_colors_for_categorical_obs(data1, groupby1)
    if groupby_colors2 not in data2.uns:
        from scanpy.plotting._utils import _set_default_colors_for_categorical_obs
        _set_default_colors_for_categorical_obs(data2, groupby2)

    fig, grid = __panel(ncols, len(plot_list), widthspace=widthspace, heightspace=heightspace)

    for count, (alignment, genename) in enumerate(plot_list):
        ax = fig.add_subplot(grid[count])
        dtw_dic = aligned_data.alignmentdict[alignment]
        path = dtw_dic[genename]["path"]
        ordered_cells1 = dtw_dic[genename]["ordered_cells1"]
        ordered_cells2 = dtw_dic[genename]["ordered_cells2"]

        ax.set_title("{}_{}".format(alignment, genename), fontsize=fontsize)
        ax.set_xlabel("Pseudotime", fontsize=fontsize)
        ax.set_aspect(0.8)
        ax.tick_params(
            labelbottom=True,
            labelleft=False,
            labelright=False,
            labeltop=False,
            bottom=False,
            left=False,
            right=False,
            top=False,
            labelsize=ticksize
            )
        ax.grid(False)

        y1 = data1[ordered_cells1, :].obs["{}_dpt_pseudotime".format(alignment)]
        y2 = data2[ordered_cells2, :].obs["{}_dpt_pseudotime".format(alignment)]

        clusters1 = data1[ordered_cells1, :].obs[groupby1]
        colors1 = data1[ordered_cells1, :].uns[groupby_colors1]
        clusters2 = data2[ordered_cells2, :].obs[groupby2]
        colors2 = data2[ordered_cells2, :].uns[groupby_colors2]

        if not type(colors1) == np.ndarray:
            colors1 = np.array([colors1])

        if not type(colors2) == np.ndarray:
            colors2 = np.array([colors2])

        dic1 = dict(
            zip(data1[ordered_cells1, :].obs[groupby1].cat.categories, colors1))
        dic2 = dict(
            zip(data2[ordered_cells2, :].obs[groupby2].cat.categories, colors2))

        colorlist1 = clusters1.replace(dic1)
        colorlist2 = clusters2.replace(dic2)

        ax.scatter(
            np.array(list(y1)), np.ones(len(y1)),
            color=colorlist1,
            zorder=-2,
        )

        ax.scatter(
            np.array(list(y2)), np.zeros(len(y2)),
            color=colorlist2,
            zorder=-2,
        )

        n_col1 = np.ceil(len(dic1.keys()) / 5).astype(int)
        ordered_clusters1 = [
            cluster for cluster in
            aligned_data.alignmentdict[alignment]['data1']
            if cluster != "#"
            ]
        patches1 = []
        for cluster in ordered_clusters1:
            patches1.append(Line2D(
                                range(1), range(1),
                                marker='o', color=dic1[cluster],
                                label=cluster, linewidth=0
                                )
                            )
        legend1 = ax.legend(handles=patches1,
                            labels=ordered_clusters1,
                            bbox_to_anchor=(1.05, 0.5),
                            loc='lower left',
                            ncol=n_col1,
                            title="{}".format(data1_name),
                            title_fontsize=legend_fontsize
                            )

        n_col2 = np.ceil(len(dic2.keys()) / 5).astype(int)
        ordered_clusters2 = [
            cluster for cluster in
            aligned_data.alignmentdict[alignment]['data2']
            if cluster != "#"
        ]
        patches2 = []
        for cluster in ordered_clusters2:
            patches2.append(Line2D(range(1), range(1),
                            marker='o', color=dic2[cluster], label=cluster, linewidth=0)
                            )
        legend2 = ax.legend(handles=patches2,
                            labels=ordered_clusters2,
                            bbox_to_anchor=(1.05, 0.5),
                            loc='upper left',
                            ncol=n_col2,
                            title="{}".format(data2_name),
                            title_fontsize=legend_fontsize
                            )

        ax.add_artist(legend1)
        ax.add_artist(legend2)

        for i, j in path:
            i = int(i)
            j = int(j)
            ax.plot((y1[i], y2[j]), (1, 0),
                    color=linecolor, alpha=0.5, zorder=-3)

        ax.set_rasterization_zorder(-1)

    if save:
        if isinstance(save, str):
            save_dir = save
            os.makedirs(os.path.dirname(save_dir), exist_ok=True)
        elif save is True:
            os.makedirs("./figures", exist_ok=True)
            save_dir = "./figures/dtw.png"
        fig.savefig(save_dir, bbox_inches='tight', pad_inches=0.1, dpi=dpi)
    if show:
        plt.show()
        plt.close()
    else:
        plt.close()
        return fig

def gene_expression_trend(
    aligned_data: CapitalData,
    gene: Union[str, list],
    alignment: Union[str, list, None] = None,
    outliers: list = [100, 0],
    polyfit_dimension: int = 3,
    switch_psedotime: bool = False,
    multi_genes: bool = True,
    data1_name: Optional[str] = "data1",
    data2_name: Optional[str] = "data2",
    data1_color: str = "tomato",
    data2_color: str = "lightskyblue",
    data1_line_color: str = "red",
    data2_line_color: str = "blue",
    ncols: int = 2,
    widthspace: float = 0.5,
    heightspace: float = 0.30,
    fontsize: float = 12,
    legend_fontsize: float = 12,
    ticksize: float = 12,
    dpi: int = 600,
    show: bool = True,
    save: Union[str, bool] = False,
):
    """\
    Plot gene expression trend.

    Parameters
    ----------
    aligned_data : CapitalData
        The data matrices containing the results of CAPTIAL.
    gene : Union[str, list]
        Keys for annotations of genes.
    alignment : Union[str, list, None], optional
        Keys for alignments to be plotted. If `None`, all alignments will be plotted, by default `None`.
    outliers : list
        Outliers for gene expression, by default [100, 0].
    polyfit_dimension : int
        Degree of the fitting polynomial, by default 3.
    switch_psedotime : bool, optional
        If `False`, data1's pseudotime is used to plot two expression trends. If `True`, switch it to data2's pseudotime, by default `False`.
    multi_genes : bool
        If `True`, figures are plotted using the cell-cell match of cp.tl.dtw(multi_genes=True), by default `False`.
    data1_name : Optional[str]
        Text of data1's legend, by default "data1".
    data2_name : Optional[str]
        Text of data2's legend, by default "data2".
    data1_color: str
        Color of data1 scatter plot, by default "tomato".
    data2_color: str
        Color of data2 scatter plot, by default "lightskyblue".
    data1_line_color: str
        Color of data1 line that are calculated using polynomial curve fitting, by default "red".
    data2_line_color: str
        Color of data2 line that are calculated using polynomial curve fitting, by default "blue".
    ncols : int
        Number of panels per row, by default 2.
    widthspace : float
        Width of space in the panels, by default 0.5.
    heightspace : float
        Height of space in the panels, by default 0.30.
    fontsize : float
        Font size of the title, by default 12.
    legend_fontsize : float, optional
        Font size of the legend, by default 12.
    ticksize : float, optional
        Tick size of the axes, by default 12.
    show : bool
        If `True`, show the figure. If `False`, return the figure, by default `True`.
    save : Union[str, bool]
        If `True` or `str`, save the figure. If a path is specified as `str`, the figure is saved in the path, by default `None`.
    """

    if not isinstance(aligned_data, CapitalData):
        ValueError(
            "draw_gene_expression_trend() expects an CapitalData argument.")

    plot_list = __set_plot_list(aligned_data, alignment, gene)

    data1 = aligned_data.adata1
    data2 = aligned_data.adata2

    fig, grid = __panel(ncols, len(plot_list), widthspace=widthspace, heightspace=heightspace)

    max_out, min_out = outliers[0], outliers[1]
    Num_quantiles = 10
    count = 0
    for count, (alignment, genename) in enumerate(plot_list):
        ax = fig.add_subplot(grid[count])
        dtw_dic = aligned_data.alignmentdict[alignment]
        if multi_genes:
            path = dtw_dic["multi_genes"]["path"]
            ordered_cells1 = dtw_dic["multi_genes"]["ordered_cells1"]
            ordered_cells2 = dtw_dic["multi_genes"]["ordered_cells2"]
        else:
            path = dtw_dic[genename]["path"]
            ordered_cells1 = dtw_dic[genename]["ordered_cells1"]
            ordered_cells2 = dtw_dic[genename]["ordered_cells2"]
            
        ax.set_title("{}, {}".format(alignment, genename))
        ax.set_xlabel("Pseudotime",fontsize=fontsize)
        ax.set_ylabel("Expression level", fontsize=fontsize)
        ax.tick_params(labelsize=ticksize)

        pseudotime = []
        data1_expression_level = []
        data2_expression_level = []

        expression1 = data1.raw.to_adata(
        )[ordered_cells1, genename].X.T[0]
        expression2 = data2.raw.to_adata(
        )[ordered_cells2, genename].X.T[0]

        if switch_psedotime:
            pseudotime = data2[ordered_cells2, :].obs["{}_dpt_pseudotime".format(
                alignment)][[j for _, j in path]].values
        else:
            pseudotime = data1[ordered_cells1, :].obs["{}_dpt_pseudotime".format(
                alignment)][[i for i, _ in path]].values

        data1_expression_level = expression1[[i for i, _ in path]]
        data2_expression_level = expression2[[j for _, j in path]]

        array = np.array(
            [pseudotime, data1_expression_level, data2_expression_level])

        # when all or most of the gene expressions are 0, np.percentile causes error
        # catching error by try and except, but it needs to be fixed
        qcut_index = pd.qcut(
            array[0], Num_quantiles, labels=False, duplicates='drop')

        gene_expression1_list = []
        for i in range(Num_quantiles):
            qcut_array = array[:, qcut_index == i]
            try:
                min_outlier, max_outlier = np.percentile(
                    qcut_array[1], q=[min_out, max_out]
                )
                gene_expression1 = qcut_array[:, (qcut_array[1] <= max_outlier) & (
                    qcut_array[1] >= min_outlier)][[0, 1]]
            except IndexError as e:
                gene_expression1 = qcut_array[[0, 1]]
            gene_expression1_list.append(gene_expression1)

        array1_poly1d = np.concatenate(gene_expression1_list, axis=1)

        gene_expression2_list = []
        for i in range(Num_quantiles):
            qcut_array = array[:, qcut_index == i]
            try:
                min_outlier, max_outlier = np.percentile(
                    qcut_array[2], q=[min_out, max_out]
                )
                gene_expression2 = qcut_array[:, (qcut_array[2] <= max_outlier) & (
                    qcut_array[2] >= min_outlier)][[0, 2]]
            except IndexError as e:
                gene_expression2 = qcut_array[[0, 2]]
            gene_expression2_list.append(gene_expression2)

        array2_poly1d = np.concatenate(gene_expression2_list, axis=1)

        y1 = np.poly1d(np.polyfit(array1_poly1d[0], array1_poly1d[1], polyfit_dimension))(
            array1_poly1d[0])
        y2 = np.poly1d(np.polyfit(array2_poly1d[0], array2_poly1d[1], polyfit_dimension))(
            array2_poly1d[0])

        ax.scatter(array[0], array[1],
                    color=data1_color, alpha=0.5, rasterized=True)
        ax.scatter(array[0], array[2],
                    color=data2_color, alpha=0.5, rasterized=True)

        if data1_name is None:
            ax.plot(array1_poly1d[0], y1, color=data1_line_color)
        else:
            ax.plot(array1_poly1d[0], y1,
                    color=data1_line_color, label=data1_name)

        if data2_name is None:
            ax.plot(array2_poly1d[0], y2, color=data2_line_color)
        else:
            ax.plot(array2_poly1d[0], y2,
                    color=data2_line_color, label=data2_name)

        if data1_name is None and data2_name is None:
            pass
        else:
            ax.legend(
                bbox_to_anchor=(1.05, 1),
                loc='upper left',
                borderaxespad=0,
                fontsize=12
                )

    if save:
        if isinstance(save, str):
            save_dir = save
            os.makedirs(os.path.dirname(save_dir), exist_ok=True)
        elif save is True:
            os.makedirs("./figures", exist_ok=True)
            save_dir = "./figures/expression_trend.png"
        fig.savefig(save_dir,  bbox_inches='tight', pad_inches=0.1, dpi=dpi)
    if show:
        plt.show()
        plt.close()
    else:
        plt.close()
        return fig


def trajectory_tree(
    adata: AnnData,
    figsize: tuple = (10, 8),
    node_size: int = 1200,
    font_size: int = 18,
    show: bool = True,
    dpi: int = 600,
    save: Union[str, bool] = False,
):
    """\
    Plot a trajectory tree.

    Parameters
    ----------
    adata : AnnData
        The annotated data matrix.
    figsize : tuple,
        Size of figure, by default (10, 8).
    node_size : int,
        Size of nodes, by default 1200.
    font_size : int,
        Font size, by default 18.
    show : bool,
        If `True`, show the figure. If `False`, return figure, by default `True`.
    save : Union[str, bool]
        If `True` or `str`, save the figure. If a path is specified as `str`, the figure is saved in the path, by default `False`.
    """

    if not isinstance(adata, AnnData):
        ValueError("draw_tree() expects an AnnData argument.")

    groupby = adata.uns["capital"]["tree"]["annotation"]
    groupby_colors = groupby + "_colors"
    if groupby_colors not in adata.uns:
        from scanpy.plotting._utils import _set_default_colors_for_categorical_obs
        _set_default_colors_for_categorical_obs(adata, groupby)

    tree = nx.convert_matrix.from_pandas_adjacency(
        adata.uns["capital"]["tree"]["tree"], create_using=nx.DiGraph)

    clusters = list(adata.obs[groupby].cat.categories)
    dic = dict(zip(clusters, adata.uns[groupby_colors]))
    colorlist = list(pd.Series(list(tree.nodes())).replace(dic))

    plt.figure(figsize=figsize)
    pos = graphviz_layout(tree, prog='dot')
    nx.draw(
        tree,
        pos,
        node_color=colorlist,
        font_weight='bold',
        font_size=font_size,
        node_size=node_size,
        with_labels=True,
        edge_color="gray",
        arrows=True
        )

    if save:
        if isinstance(save, str):
            save_dir = save
            os.makedirs(os.path.dirname(save_dir), exist_ok=True)
        elif save is True:
            os.makedirs("./figures", exist_ok=True)
            save_dir = "./figures/tree.png"
        plt.savefig(save_dir,  bbox_inches='tight', pad_inches=0.1, dpi=dpi)
    if show:
        plt.show()
        plt.close()
    else:
        plt.close()



def tree_alignment(
    aligned_data: CapitalData,
    figsize: tuple = (10, 8),
    node_size: int = 1200,
    font_size: int = 18,
    dpi: int = 600,
    show: bool = True,
    save: Union[str, bool] = False,
):
    """\
    Plot the alignment of the trees.

    Parameters
    ----------
    aligned_data : CapitalData
        [description]
    figsize : tuple
        [description], by default (10, 8).
    node_size : int
        [description], by default 1200.
    font_size : int
        [description], by default 18.
    show : bool,
        If `True`, show the figure. If `False`, return figure, by default `True`.
    save : Union[str, bool]
        If `True` or `str`, save the figure. If a path is specified as `str`, the figure is saved in the path, by default `None`.
    """

    if not isinstance(aligned_data, CapitalData):
        ValueError("draw_alignmenttree() expects an CapitalData argument.")

    tree = aligned_data.alignedtree
    mapping = {i: str(i).replace("'", "").replace(
        "(", "").replace(")", "") for i in tree.nodes()}
    tree = nx.relabel_nodes(tree, mapping)
    plt.figure(figsize=figsize)
    pos = graphviz_layout(tree, prog='dot')
    nx.draw(tree, pos, node_color='lightskyblue', font_weight='bold', font_size=font_size,
            node_size=node_size, with_labels=True, edge_color="gray", arrows=True)

    if save:
        if isinstance(save, str):
            save_dir = save
            os.makedirs(os.path.dirname(save_dir), exist_ok=True)
        elif save is True:
            os.makedirs("./figures", exist_ok=True)
            save_dir = "./figures/alignmenttree.png"
        plt.savefig(save_dir,  bbox_inches='tight', pad_inches=0.1,dpi = dpi)
    if show:
        plt.show()
    else:
        plt.close()
    plt.close()
    
    
def __panel(
    ncols,
    num_panels,
    widthspace=0.15,
    heightspace=0.30
):
    from matplotlib import gridspec

    n_panels_x = min(ncols, num_panels)
    n_panels_y = np.ceil(num_panels / n_panels_x).astype(int)
    fig = plt.figure(
        figsize=(
            n_panels_x * rcParams['figure.figsize'][0] * (1.20),
            n_panels_y * rcParams['figure.figsize'][1],
        ),
    )
    grid = gridspec.GridSpec(
        nrows=n_panels_y,
        ncols=n_panels_x,
        hspace=heightspace,
        wspace=widthspace,
    )
    return fig, grid


def __set_plot_list(
    aligned_data,
    alignment,
    genes,
    multi_genes = False
):
    alignmentlist = []
    if alignment is None:
        alignmentlist = list(aligned_data.alignmentdict.keys())
    else:
        if isinstance(alignment, list):
            alignmentlist = alignment
        elif isinstance(alignment, str):
            alignmentlist = [alignment]
        else:
            raise ValueError("alignment must be list or str of alignment \
            e.g 'alignment000' or ['alignment000','alignment001', ...].")

    alignment_check = [
        alignment for alignment in alignmentlist
        if alignment not in aligned_data.alignmentdict.keys()
    ]
    if len(alignment_check) > 0:
        raise ValueError(
            "Alignments were not found. {}".format(alignment_check)
        )

    if isinstance(genes, list):
        genenamelist = genes
    elif isinstance(genes, str):
        genenamelist = [genes]
    elif isinstance(genes, np.ndarray):
        genenamelist = list(genes)
    else:
        raise ValueError("gene must be list, str or np.ndarray.")

    plot_list_tmp = list(itertools.product(alignmentlist, genenamelist))
    plot_list = [
        (alignment, gene) for alignment, gene in plot_list_tmp
        if gene in aligned_data.alignmentdict[alignment] or multi_genes
    ]

    if len(plot_list) == 0:
        raise ValueError(
            "Genes were not found in alignments. Run capital.dtw() first.")
        
    if multi_genes:
        pass
    else:
        non_plot_list = [
            gene + " not in " + alignment for alignment, gene in plot_list_tmp
            if gene not in aligned_data.alignmentdict[alignment] 
        ]

        if len(non_plot_list) > 0:
            print("The genes below were not found in the alignments. Run capital.dtw() or specify the alignments to draw.")
            if len(non_plot_list) < 5:
                print(non_plot_list)
            else:
                print("{} and more...".format(non_plot_list[:5]))
    
    return plot_list


import numpy as np
import scanpy as sc
from scipy.spatial.distance import pdist, squareform
from tslearn.metrics import dtw_path, dtw
from sklearn.preprocessing import MinMaxScaler
from .._util import CapitalData


class DPT:
    def __init__(self):
        pass

    def _calc_root_cell(
        self,
        adata,
    ):
        if not isinstance(adata.X, np.ndarray):
            X = adata.X.toarray()
        
        else:
            X = adata.X

        root_cluster = adata.uns['capital']['tree']['root_node']
        groupby = adata.uns["capital"]["tree"]["annotation"]

        # Choose the root cell
        Y = pdist(X, "euclidean")
        distancearray = squareform(Y)

        distancearray = np.sum(distancearray, axis=1)
        loc = np.argsort(distancearray)
        count = 0

        while adata.obs[groupby][np.where(loc == count)[0][0]] != root_cluster:
            count += 1

        # Store the root cell id
        root_cell = adata.obs.index[np.flatnonzero(adata.obs[groupby])[np.where(loc == count)[0][0]]]
        adata.uns["root_cell"] = root_cell

    def _dpt_for_an_alignment(
        self,
        adata,
        cluster_list,
        alignment_id,
        copy=False
    ):

        adata = adata.copy() if copy else adata
        groupby = adata.uns["capital"]["tree"]["annotation"]
        root_cluster = cluster_list[0]
        # get  cells only in the clusters of cluster_list
        adata_dpt = adata[adata.obs[groupby].isin(
            cluster_list)].copy()

        # Choose cells only in "clusters"
        adata_dpt = adata[adata.obs[groupby].isin(cluster_list)].copy()

        # "iroot" is a cell that is the source
        adata_dpt.uns["iroot"] = adata_dpt.obs.index.get_loc(adata_dpt.uns["root_cell"])

        # process diffusion maps and dpt, calculate "dpt_pseudotime"
        sc.pp.neighbors(adata_dpt)
        sc.tl.diffmap(adata_dpt)
        sc.tl.dpt(adata_dpt)

        adata.obs["{}_dpt_pseudotime".format(
            alignment_id)] = adata_dpt.obs["dpt_pseudotime"]
        adata.uns["capital"]["pseudotime"]["{}".format(alignment_id)] = {}

        each_pseudotime_dict = adata.uns["capital"]["pseudotime"]["{}".format(
            alignment_id)]
        each_pseudotime_dict["clusters"] = np.array(cluster_list, dtype=object)
        # add dpt_pseudotime for the clusters in one alignment to adata.obs["alignment000_dpt_pseudotime"]
        # add clusters name used in the alignment  to adata.uns["capital"]["pseudotime"]["alignment000"]["clusters"]
        return adata if copy else None

    def dpt_for_alignments(
        self,
        aligned_data: CapitalData,
        alignment=None,
        no_prune=False
    ):
        groupby1 = aligned_data.adata1.uns["capital"]["tree"]["annotation"]
        groupby2 = aligned_data.adata2.uns["capital"]["tree"]["annotation"]
        aligned_data.adata1.uns["capital"]["pseudotime"] = {}
        aligned_data.adata2.uns["capital"]["pseudotime"] = {}

        alignment_id_list = []
        if alignment is None:
            alignment_id_list = list(aligned_data.alignmentdict.keys())
        else:
            if isinstance(alignment, list):
                alignment_id_list = alignment
            elif isinstance(alignment, str):
                alignment_id_list = [alignment]
            else:
                raise ValueError(
                    "alignment must be list or str of alignment. "
                    "e.g. 'alignment000' or ['alignment000','alignment001', ...].")

        self._calc_root_cell(aligned_data.adata1)
        self._calc_root_cell(aligned_data.adata2)

        for alignment_id in alignment_id_list:
            route1 = aligned_data.alignmentdict[alignment_id]["data1"]
            route2 = aligned_data.alignmentdict[alignment_id]["data2"]
            if all([i == "#" for i in route1]):
                continue
            if all([i == "#" for i in route2]):
                continue

            if not no_prune:
                tmp = []
                for i in range(len(route1)):
                    if route1[i] == "#" or route2[i] == "#":
                        tmp.append(i)
                    if route1[i] != "#" and route2[i] != "#":
                        break
                for i in reversed(range(len(route1))):
                    if route1[i] == "#" or route2[i] == "#":
                        tmp.append(i)
                    if route1[i] != "#" and route2[i] != "#":
                        break
                cluster_list1 = [route1[i]
                                for i in list(range(len(route1))) if i not in tmp]
                cluster_list2 = [route2[i]
                                for i in list(range(len(route2))) if i not in tmp]

            cluster_list1 = [
                node for node in route1
                if node in aligned_data.adata1.obs[groupby1].values
            ]
            cluster_list2 = [
                node for node in route2
                if node in aligned_data.adata2.obs[groupby2].values
            ]

            if len(cluster_list1) == 0 or len(cluster_list2) == 0:
                break

            self._dpt_for_an_alignment(
                aligned_data.adata1, cluster_list1, alignment_id)
            self._dpt_for_an_alignment(
                aligned_data.adata2, cluster_list2, alignment_id)



class DynamicTimeWarping():
    def __init__(self):
        pass

    def dtw_for_alignments(
        self,
        aligned_data: CapitalData,
        gene,
        alignment=None,
        multi_genes=False
    ):
        groupby1 = aligned_data.adata1.uns["capital"]["tree"]["annotation"]
        groupby2 = aligned_data.adata2.uns["capital"]["tree"]["annotation"]

        alignment_id_list = []
        if alignment is None:
            alignment_id_list = list(aligned_data.alignmentdict.keys())
        else:
            if isinstance(alignment, list):
                alignment_id_list = alignment
            elif isinstance(alignment, str):
                alignment_id_list = [alignment]
            else:
                raise ValueError(
                    "alignment must be list or str of alignment. "\
                    "e.g. 'alignment000' or ['alignment000','alignment001', ...].")

        if isinstance(gene, list):
            genenamelist = gene
        elif isinstance(gene, str):
            genenamelist = [gene]
        elif isinstance(gene, np.ndarray):
            genenamelist = list(gene)
        else:
            raise ValueError("gene must be list, str or np.ndarray.")

        for alignment_id in alignment_id_list:
            cluster_list1 = aligned_data.alignmentdict[alignment_id]["data1"]
            cluster_list2 = aligned_data.alignmentdict[alignment_id]["data2"]

            adata_dpt1 = aligned_data.adata1[aligned_data.adata1.obs[groupby1].isin(
                cluster_list1)].copy()
            adata_dpt1 = adata_dpt1[adata_dpt1.obs.sort_values(
                "{}_dpt_pseudotime".format(alignment_id)).index].copy()
            adata_dpt2 = aligned_data.adata2[aligned_data.adata2.obs[groupby2].isin(
                cluster_list2)].copy()
            adata_dpt2 = adata_dpt2[adata_dpt2.obs.sort_values(
                "{}_dpt_pseudotime".format(alignment_id)).index].copy()

            if multi_genes:
                ordered_cells1, ordered_cells2, path, dist = self._applying_dtw_to_clusters(
                    adata_dpt1, adata_dpt2, gene)
                result = aligned_data.alignmentdict[alignment_id]
                result["multi_genes"] = {"ordered_cells1": ordered_cells1,
                                    "ordered_cells2": ordered_cells2,
                                    "path": path,
                                    }
            else:
                for genename in genenamelist:
                    ordered_cells1, ordered_cells2, path, dist = self._applying_dtw_to_clusters(
                        adata_dpt1, adata_dpt2, genename)
                    result = aligned_data.alignmentdict[alignment_id]
                    result[genename] = {"ordered_cells1": ordered_cells1,
                                        "ordered_cells2": ordered_cells2,
                                        "path": path,
                                        }

    # data used to do dynamic time warping are stored in
    # file1_ordered_data, file2_ordered_data, paths
    # if nodes tha are compared are empty, data are stored as "#"
    # if the opponent side of the node is empty, the data is sorted and stored,
    # but the another node and path are stored as "#"

    # When using MAGIC there aren't many outlier or dropout,
    # so outlier aren't taken any
    # expecting Anndata with one gene

    def _applying_dtw_to_clusters(
        self,
        adata1,
        adata2,
        genename,
        min_percentile_outlier=0,
        max_percentile_outlier=100
    ):
        expression1 = adata1.raw.to_adata()[:, genename].X
        expression2 = adata2.raw.to_adata()[:, genename].X

        if not isinstance(expression1, np.ndarray):
            expression1 = expression1.toarray()
        if not isinstance(expression2, np.ndarray):
            expression2 = expression2.toarray()

        # when all or most of the gene expressions are 0, np.percentile causes error
        # catching error by try and except, but it needs to be fixed
        # excluding cells that have too low or too high gene expression
        try:
            min_outlier, max_outlier = np.percentile(
                expression1, q=[min_percentile_outlier, max_percentile_outlier]
            )
            gene_expression1 = expression1[(expression1 <= max_outlier) & (
                expression1 >= min_outlier)]
            ordered_cells1 = np.array(adata1[(
                expression1 <= max_outlier) & (expression1 >= min_outlier)].obs_names.to_list())
        except IndexError as e:
            gene_expression1 = expression1
            ordered_cells1 = np.array(adata1.obs_names.to_list())

        # excluding cells that have too low or too high gene expression
        try:
            min_outlier, max_outlier = np.percentile(
                expression2, q=[min_percentile_outlier, max_percentile_outlier]
            )

            gene_expression2 = expression2[(
                expression2 <= max_outlier) & (expression2 >= min_outlier)]

            ordered_cells2 = np.array(adata2[(
                expression2 <= max_outlier) & (expression2 >= min_outlier)].obs_names.to_list())
        except IndexError as e:
            gene_expression2 = expression2
            ordered_cells2 = np.array(adata2.obs_names.to_list())

        path, dist = dtw_path(
            gene_expression1, gene_expression2)

        return ordered_cells1, ordered_cells2, path, dist

    def get_genes_similarity_score(
        self,
        aligned_data: CapitalData,
        gene=None,
        alignment=None,
        min_disp=1.0,
    ):

        groupby1 = aligned_data.adata1.uns["capital"]["tree"]["annotation"]
        groupby2 = aligned_data.adata2.uns["capital"]["tree"]["annotation"]

        alignment_id_list = []
        if alignment is None:
            alignment_id_list = list(aligned_data.alignmentdict.keys())
        else:
            if isinstance(alignment, list):
                alignment_id_list = alignment
            elif isinstance(alignment, str):
                alignment_id_list = [alignment]
            else:
                raise ValueError(
                    "alignment must be list or str of alignment. "\
                    "e.g. 'alignment000' or ['alignment000','alignment001', ...].")

        genenamelist = []
        if gene is not None:
            if isinstance(gene, list):
                genenamelist = gene
            elif isinstance(gene, str):
                genenamelist = [gene]
            elif isinstance(gene, np.ndarray):
                genenamelist = list(gene)
            else:
                raise ValueError("gene must be list, str or np.ndarray.")


        if aligned_data.similarity_score is None:
            dic_similarity_score = {}
        else:
            dic_similarity_score = aligned_data.similarity_score

        for alignment_id in alignment_id_list:
            cluster_list1 = aligned_data.alignmentdict[alignment_id]["data1"]
            cluster_list2 = aligned_data.alignmentdict[alignment_id]["data2"]

            adata_dpt1 = aligned_data.adata1[aligned_data.adata1.obs[groupby1].isin(
                cluster_list1)].copy()
            adata_dpt1 = adata_dpt1[adata_dpt1.obs.sort_values(
                "{}_dpt_pseudotime".format(alignment_id)).index].raw.to_adata()
            adata_dpt2 = aligned_data.adata2[aligned_data.adata2.obs[groupby2].isin(
                cluster_list2)].copy()
            adata_dpt2 = adata_dpt2[adata_dpt2.obs.sort_values(
                "{}_dpt_pseudotime".format(alignment_id)).index].raw.to_adata()

            if gene is None:
                sc.pp.highly_variable_genes(adata_dpt1)
                sc.pp.highly_variable_genes(adata_dpt2)
                s1 = set(adata_dpt1.var.index)
                s2 = set(adata_dpt2.var.index)
                genenamelist = list(s1.intersection(s2))
                disp1 = adata_dpt1[:, genenamelist].var["dispersions_norm"]
                disp2 = adata_dpt2[:, genenamelist].var["dispersions_norm"]

                genenamelist = disp1[(
                    disp1 > min_disp) | (disp2 > min_disp)].index.values

            print("Calculating similarity score of {} genes in {}".format(
                len(genenamelist), alignment_id))

            score_list = self._get_dtw_score(
                adata_dpt1, adata_dpt2, genenamelist)
            ar = np.array([genenamelist, score_list], dtype=object)
            ar = ar[:, np.argsort(ar[1])][0]
            dic_similarity_score[alignment_id] = ar

        aligned_data.similarity_score = dic_similarity_score
        print("Calculating finished")

    def _get_dtw_score(
        self,
        adata1,
        adata2,
        genenamelist,
    ):
        dist_list = []
        expression1 = adata1[:, genenamelist].X
        expression2 = adata2[:, genenamelist].X

        if not isinstance(expression1, np.ndarray):
            expression1 = expression1.toarray()
        if not isinstance(expression2, np.ndarray):
            expression2 = expression2.toarray()

        mmscaler = MinMaxScaler()
        expression1 = mmscaler.fit_transform(expression1)
        expression2 = mmscaler.fit_transform(expression2)

        for i in range(len(genenamelist)):
            dist = dtw(
                expression1[:, i], expression2[:, i])
            dist_list.append(dist)

        return dist_list



import itertools
from collections import deque
import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc

from .._util import CapitalData


class Tree_Alignment:
    def __init__(self):
        self.__successors1 = None
        self.__postorder1 = None
        self.__tree1 = None
        self.__successors2 = None
        self.__postorder2 = None
        self.__tree2 = None
        self.__forestdistance = None
        self.__traceforest = None
        self.__treedistance = None
        self.__tracetree = None
        self.__alignmentcost = None

    def tree_alignment(
        self,
        adata1,
        adata2,
        cost=1.0,
        N_1=2000,
        N_2=2000
    ):

        COST = cost
        gene_list = self.sort_data(
            adata1, adata2, N_1, N_2)

        adata1.uns["capital"]["intersection_genes"] = np.array(
            gene_list, dtype=object)
        adata2.uns["capital"]["intersection_genes"] = np.array(
            gene_list, dtype=object)

        self._dp(adata1, adata2, gene_list, COST)
        alignedtree = self._traceback()

        path_cluster_list = []
        source_node = list(nx.topological_sort(alignedtree))[0]
        for node in list(alignedtree.nodes):
            if alignedtree.out_degree(node) == 0:
                cluster_list = nx.shortest_path(
                    alignedtree, source=source_node, target=node)
                route1 = [i[0] for i in cluster_list]
                route2 = [i[1] for i in cluster_list]
                path_cluster_list.append([route1, route2])

        alignmentdict = {"alignment{:03d}".format(i):
                            {"data1": clusters[0],
                            "data2": clusters[1]}
                            for i, clusters in enumerate(path_cluster_list)}

        aligned_data = CapitalData(
            adata1.copy(),
            adata2.copy(),
            alignedtree,
            np.array([self.__alignmentcost], dtype=float),
            np.array(gene_list, dtype=object),
            alignmentdict,
        )

        return aligned_data

    def _set_initial_condition(
        self,
        data1,
        data2,
        cost=1.0
    ):

        self.__successors1 = data1.uns["capital"]["tree"]["successors"]
        self.__postorder1 = data1.uns["capital"]["tree"]["postorder"]
        self.__tree1 = nx.convert_matrix.from_pandas_adjacency(
            data1.uns["capital"]["tree"]["tree"], create_using=nx.DiGraph)
        self.__successors2 = data2.uns["capital"]["tree"]["successors"]
        self.__postorder2 = data2.uns["capital"]["tree"]["postorder"]
        self.__tree2 = nx.convert_matrix.from_pandas_adjacency(
            data2.uns["capital"]["tree"]["tree"],create_using=nx.DiGraph)

        # get combination of children
        # D(F1[i],F2[j]) is stored in forestdistance.loc[i,j]
        # D(F1[i1,i2],F2[j]) is stored in forestdistance.loc["(i1,i2)",j]
        # D({T1[i]},F2[j]) is stored in forestdistance.loc["(i,)", j]
        forest1_combinations = []
        for child in self.__successors1.values():
            if child.size == 1:
                children = list(itertools.combinations(child, 1))
                forest1_combinations.extend(children)
            elif child.size >= 1:
                for k in range(1, child.size):
                    children = list(itertools.combinations(child, k))
                    forest1_combinations.extend(children)

        forest2_combinations = []
        for child in self.__successors2.values():
            if child.size == 1:
                children = list(itertools.combinations(child, 1))
                forest2_combinations.extend(children)
            elif child.size >= 1:
                for k in range(1, child.size):
                    children = list(itertools.combinations(child, k))
                    forest2_combinations.extend(children)

        forest1 = [i for i in list(self.__tree1.nodes)] + \
            forest1_combinations + ["#"]
        forest2 = [j for j in list(self.__tree2.nodes)] + \
            forest2_combinations + ["#"]
        forest1 = list(map(str, forest1))
        forest2 = list(map(str, forest2))
        forest = pd.DataFrame(index=forest1, columns=forest2)
        forest.loc["#", "#"] = 0

        tree = pd.DataFrame(
            index=list(map(str, list(self.__tree1))) + ["#"],
            columns=list(map(str, list(self.__tree2))) + ["#"])
        tree.loc["#", "#"] = 0

        self.__forestdistance = forest
        self.__traceforest = pd.DataFrame(index=forest1, columns=forest2)
        self.__treedistance = tree
        self.__tracetree = pd.DataFrame(
            index=list(map(str, list(self.__tree1))) + ["#"],
            columns=list(map(str, list(self.__tree2))) + ["#"])

        COST = cost

        for i in self.__postorder1:
            size, successors = self._get_successors(self.__successors1, i)
            if size == 1:
                # D(F1[i],θ) = Σ D(T1[ik],θ)
                self._setF(i, "#", self._getT(successors[0], "#"))
                self._setT(i, "#", self._getF(i, "#") + COST)
            else:
                # D(F1[i],θ) = Σ D(T1[ik],θ)
                tmp = 0
                for ichild in successors:
                    tmp += self._getT(ichild, "#")
                self._setF(i, "#", tmp)

                # D({T1[ip],...,T1[iq]},θ) = D(T1[ip],θ) + ... + D(T1[iq],θ)
                for k in range(1, size):
                    children = list(itertools.combinations(successors, k))

                    for ichild in children:
                        tmp = 0
                        for k in ichild:
                            tmp += self._getT(k, "#")
                        self._setF(ichild, "#", tmp)
                self._setT(i, "#", self._getF(i, "#") + COST)

        for j in self.__postorder2:
            size, successors = self._get_successors(self.__successors2, j)
            if size == 1:
                self._setF("#", j, self._getT("#", successors[0]))
                self._setT("#", j, self._getF("#", j) + COST)
            else:
                tmp = 0
                for jchild in successors:
                    tmp += self._getT("#", jchild)
                self._setF("#", j, tmp)

                for k in range(1, size):
                    children = list(itertools.combinations(successors, k))

                    for jchild in children:
                        tmp = 0
                        for k in jchild:
                            tmp += self._getT("#", k)
                        self._setF("#", jchild, tmp)
                self._setT("#", j, self._getF("#", j) + COST)

    @property
    def forestdistance(self):
        return self.__forestdistance

    @property
    def traceforest(self):
        return self.__traceforest

    @property
    def tracetree(self):
        return self.__tracetree

    @property
    def treedistance(self):
        return self.__treedistance

    def sort_data(
        self,
        adata1,
        adata2,
        N_1=None,
        N_2=None
    ):
        if N_1 is not None:
            adata1 = adata1.raw.to_adata()
            sc.pp.highly_variable_genes(adata1, n_top_genes=N_1)
            adata1 = adata1[:, adata1.var['highly_variable']]
        elif N_1 is None:
            pass

        if N_2 is not None:
            adata2 = adata2.raw.to_adata()
            sc.pp.highly_variable_genes(adata2, n_top_genes=N_2)
            adata2 = adata2[:, adata2.var['highly_variable']]
        elif N_2 is None:
            pass

        s1 = set(adata1.var.index)
        s2 = set(adata2.var.index)
        intersection_list = list(s1.intersection(s2))

        if len(intersection_list) < 2:
            raise ValueError("highly variable genes of intersection of data1 and data2 are not enough "\
                                "to calculate the cost of a tree alignment. \n"\
                                "Specify num_genes1 and num_genes2 carefully.")

        print("{} genes are used to calculate cost of tree alignment.\n".format(
            len(intersection_list)))

        return intersection_list

    # cluster_centroid: pd.DataFrame
    # index is cluster name, columns is gene name, X is gene expression level
    def _calculate_cluster_centroid_for_genes(
        self,
        adata,
        gene_list,
    ):
        groupby = adata.uns["capital"]["tree"]["annotation"]
        filtered_data = adata.raw.to_adata()[:, gene_list]
        cluster_centroid_data = np.empty((0, filtered_data.n_vars))
        clustername = filtered_data.obs[groupby].unique().tolist()

        for i in clustername:
            a_cluster_data = filtered_data[filtered_data.obs[groupby] == "{}".format(
                i)].to_df()
            a_cluster_median = a_cluster_data.median(axis=0).values
            cluster_centroid_data = np.vstack(
                (cluster_centroid_data, a_cluster_median)
            )
        cluster_centroid = pd.DataFrame(
            cluster_centroid_data,
            index=clustername,
            columns=filtered_data.var_names
        )
        return cluster_centroid
    # return length of i's children and tuple of children

    def _get_successors(self, successors, i):
        size = successors[i].size
        successor = tuple([str(k) for k in successors[i]])
        if len(successor) == 0:
            successor = ("#")

        return size, successor

    def _setF(self, i, j, distance):
        if isinstance(i, tuple):
            if len(i) == 0:
                i = "#"
            i = str(i)

        if isinstance(j, tuple):
            if len(j) == 0:
                j = "#"
            j = str(j)

        if i not in self.__forestdistance.index:
            print("Error: {} does not exist in forestdistance index.".format(i))

        if j not in self.__forestdistance.columns:
            print("Error: {} does not exist in forestdistance columns.".format(j))

        self.__forestdistance.loc[i, j] = distance

    def _settraceF(self, i, j, trace):
        if isinstance(i, tuple):
            if len(i) == 0:
                i = "#"
            i = str(i)

        if isinstance(j, tuple):
            if len(j) == 0:
                j = "#"
            j = str(j)
        if i not in self.__traceforest.index:
            print("Error: {} does not exist in traceforest index.".format(i))

        if j not in self.__traceforest.columns:
            print("Error: {} does not exist in traceforest columns.".format(j))
        self.__traceforest.loc[i, j] = trace

    def _settraceT(self, i, j, trace):
        if isinstance(i, tuple):
            if len(i) == 0:
                i = "#"
        i = str(i)

        if isinstance(j, tuple):
            if len(j) == 0:
                j = "#"
        j = str(j)

        if i not in self.__tracetree.index:
            print("Error: {} does not exist in tracetree index.".format(i))

        if j not in self.__tracetree.columns:
            print("Error: {} does not exist in tracetree columns.".format(j))

        self.__tracetree.loc[i, j] = trace

    def _setT(self, i, j, distance):
        if isinstance(i, tuple):
            if len(i) == 0:
                i = "#"
        i = str(i)

        if isinstance(j, tuple):
            if len(j) == 0:
                j = "#"
        j = str(j)

        if i not in self.__treedistance.index:
            print("Error: {} does not exist in treedistance index.".format(i))

        if j not in self.__treedistance.columns:
            print("Error: {} does not exist in treedistance columns.".format(j))

        self.__treedistance.loc[i, j] = distance

    def _getF(self, i, j, parent1="Nan", parent2="Nan"):
        if isinstance(i, tuple):
            if len(i) == 0:
                i = "#"
        i = str(i)

        if isinstance(j, tuple):
            if len(j) == 0:
                j = "#"
        j = str(j)

        if i not in self.__forestdistance.index:
            i = str(parent1)

        if j not in self.__forestdistance.columns:
            j = str(parent2)

        F = self.__forestdistance.loc[i, j]

        return F

    def _getT(self, i, j):
        i = str(i)
        j = str(j)

        if i not in self.__treedistance.index:
            print("Error: TreeDistance index called does not exist.")

        if j not in self.__treedistance.columns:
            print("Error: TreeDistance columns called does not exist.")

        T = self.__treedistance.loc[i, j]

        return T

    def _cal1(self, A, B):
        if not isinstance(A, tuple):
            _, A = self._get_successors(self.__successors1, A)

        if not isinstance(B, tuple):
            _, B = self._get_successors(self.__successors2, B)

        mintemp = 1000
        trace = "Nan"
        temp = 0

        for k in A:
            for l in B:
                Asub = tuple([i for i in A if i != k])
                Bsub = tuple([j for j in B if j != l])

                temp = self._getF(Asub, Bsub) + self._getT(k, l)

                if mintemp > temp:
                    mintemp = temp
                    if Bsub == ():
                        Bsub = "#"
                    if l == ():
                        l = "#"
                    if Asub == ():
                        Asub = "#"
                    if k == ():
                        k = "#"

                    trace = [1, [Asub, Bsub], [k, l]]

        return mintemp, trace

    def _cal2(self, A, B, cost):
        COST = cost
        parentA = "Nan"
        parentB = "Nan"
        if not isinstance(A, tuple):
            parentA = A
            _, A = self._get_successors(self.__successors1, A)

        if not isinstance(B, tuple):
            parentB = B
            _, B = self._get_successors(self.__successors2, B)

        Bprime = [()]
        for m in range(1, len(B)+1):
            for Bp in list(itertools.combinations(B, m)):
                Bprime.extend([Bp])

        mintemp = 1000
        trace = "Nan"
        temp = 0

        for k in A:
            for Bp in Bprime:
                Asub = tuple([i for i in A if i != k])
                Bsub = tuple([j for j in B if not j in Bp])
                if k == "#":
                    temp = self._getF(Asub, Bsub, parent2=parentB) + \
                        self._getF("#", Bp, parent2=parentB)
                else:
                    temp = self._getF(Asub, Bsub, parent2=parentB) + \
                        self._getF(k, Bp, parent2=parentB) + COST

                if mintemp > temp:
                    mintemp = temp
                    if Bsub == ():
                        Bsub = "#"
                    if Bp == ():
                        Bp = "#"
                    if Asub == ():
                        Asub = "#"
                    if k == ():
                        k = "#"
                    if not "{}".format(Bsub) in self.__forestdistance.columns:
                        Bsub = parentB
                    if not "{}".format(Bp) in self.__forestdistance.columns:
                        Bp = parentB

                    trace = [2, [[Asub, Bsub], [k, Bp]], (k, "#")]

        return mintemp, trace

    # min D(A-Aprime,B-{T2[jq]}) + D(Aprime,F2[jq]) + cost
    def _cal3(self, A, B, cost):
        COST = cost
        parentA = "Nan"
        parentB = "Nan"

        if not isinstance(A, tuple):
            parentA = A
            _, A = self._get_successors(self.__successors1, A)

        if not isinstance(B, tuple):
            parentB = B
            _, B = self._get_successors(self.__successors2, B)

        Aprime = [()]
        for m in range(1, len(A)+1):
            for Ap in list(itertools.combinations(A, m)):
                Aprime.extend([Ap])

        mintemp = 1000
        trace = "Nan"
        temp = 0

        for l in B:
            for Ap in Aprime:

                Asub = tuple([i for i in A if i not in Ap])
                Bsub = tuple([j for j in B if j != l])
                temp = self._getF(Asub, Bsub, parent1=parentA) + \
                    self._getF(Ap, l, parent1=parentA) + COST

                if mintemp > temp:
                    mintemp = temp
                    if Bsub == ():
                        Bsub = "#"
                    if l == ():
                        l = "#"
                    if Asub == ():
                        Asub = "#"
                    if Ap == ():
                        Ap = "#"

                    if not "{}".format(Asub) in self.__forestdistance.index:
                        Asub = parentA
                    if not "{}".format(Ap) in self.__forestdistance.index:
                        Ap = parentA
                    trace = [3, [[Asub, Bsub], [Ap, l]], ("#", l)]

        return mintemp, trace

    # trace is a list that has 3 pairs like [a,(b),(c)]
    # a is a record that show which calculation was done
    # (b) is a pair or 2 pairs that goes to the search in traceforest for the next traceback
    # (c) is a pair that goes to stack or foreststack

    def _calculateForest(
        self,
        A,
        B,
        cost
    ):
        COST = cost
        min1, trace1 = self._cal1(A, B)
        min2, trace2 = self._cal2(A, B, COST)
        min3, trace3 = self._cal3(A, B, COST)

        forestdistance = np.array([min1, min2, min3], dtype=object).min()
        trace = [trace1, trace2, trace3][np.array([min1, min2, min3], dtype=object).argmin()]

        return forestdistance, trace

    # trace is a list that has 3 pairs like [(a),(b),(c)]
    # (a) is a record that show which calculation was done
    # (b) is a pair of the match result for D(T[i],T[j])
    # (c) is a pair that goes to the stack
    def _calculateTreedistance(
        self,
        i,
        j,
        mincost
    ):
        MINCOST = mincost
        _, successor1 = self._get_successors(self.__successors1, i)
        _, successor2 = self._get_successors(self.__successors2, j)

        distancelist = []
        for l in successor2:
            distancelist.append(
                [self._getT("#", j) + self._getT(i, l) - self._getT("#", l), [0, ("#", j), (i, l)]])
        for k in successor1:
            distancelist.append(
                [self._getT(i, "#") + self._getT(k, j) - self._getT(k, "#"), [0, (i, "#"), (k, j)]])

        distancelist.append([self._getF(i, j) + MINCOST, [1, (i, j), (i, j)]])

        array = np.array(distancelist, dtype=object)

        treedistance = array[:, 0].min()
        trace = array[array[:, 0].argmin(), 1]

        return treedistance, trace

    def _dp(
        self,
        adata1,
        adata2,
        gene_list,
        cost
    ):
        # np.warning cause warning below
        # VisibleDeprecationWarning: Creating an ndarray from ragged nested sequences is deprecated.
        # If you meant to do this, you must specify 'dtype=object' when creating the ndarray.
        # however it causes inside pandas and to recover this wanrnig, we need to rewrite all the code above
        # and it still get the result we need. we will get this right soon.
        np.warnings.filterwarnings(
            'ignore', category=np.VisibleDeprecationWarning)

        COST = cost

        self._set_initial_condition(adata1, adata2, cost=COST)
        cluster_centroid1 = self._calculate_cluster_centroid_for_genes(
            adata1, gene_list)
        cluster_centroid2 = self._calculate_cluster_centroid_for_genes(
            adata2, gene_list)

        for i in self.__postorder1:
            for j in self.__postorder2:
                df = pd.DataFrame(
                    {"A": cluster_centroid1.loc[i], "B": cluster_centroid2.loc[j]})
                mincost = 1 - df.corr(method="spearman").iloc[0, 1]

                size1, successor1 = self._get_successors(self.__successors1, i)
                size2, successor2 = self._get_successors(self.__successors2, j)

                Alist = []
                if size1 == 0:
                    pass
                elif size1 == 1:
                    Alist.extend([(successor1)])
                else:
                    for m in range(1, len(successor1)):
                        for l in list(itertools.combinations(successor1, m)):
                            Alist.extend([l])

                Alist.extend([i])
                if size2 == 0:
                    pass
                elif size2 == 1:
                    for A in Alist:
                        fdistance, ftrace = self._calculateForest(
                            A, (successor2), COST)
                        self._setF(A, (successor2), fdistance)
                        self._settraceF(A, (successor2), ftrace)
                else:
                    for m in range(1, len(successor2)):
                        for B in list(itertools.combinations(successor2, m)):
                            for A in Alist:
                                fdistance, ftrace = self._calculateForest(
                                    A, B, COST)
                                self._setF(A, B, fdistance)
                                self._settraceF(A, B, ftrace)

                for A in Alist:
                    fdistance, ftrace = self._calculateForest(A, j, COST)
                    self._setF(A, j, fdistance)
                    self._settraceF(A, j, ftrace)

                tdistance, ttrace = self._calculateTreedistance(i, j, mincost)
                self._setT(i, j, tdistance)
                self._settraceT(i, j, ttrace)

    def _traceback(self):
        G = nx.DiGraph()
        G.add_node("tempnode")
        parent = "tempnode"
        stack = deque()
        stack.append(
            (self.__postorder1[-1], self.__postorder2[-1], parent))

        while len(stack) != 0:
            i, j, parent = stack.pop()

            if i == "#" and j == "#":
                continue

            elif i == "#":
                if j != "#":
                    H = nx.dfs_tree(self.__tree2, j)
                    G = nx.compose(G, H)
                    G.add_edge(parent, j)
                    G = nx.relabel_nodes(
                        G, dict(zip(list(H.nodes), [("#", k) for k in H.nodes])))
                continue

            elif j == "#":
                if i != "#":
                    H = nx.dfs_tree(self.__tree1, i)
                    G = nx.compose(G, H)
                    G.add_edge(parent, i)
                    G = nx.relabel_nodes(
                        G, dict(zip(list(H.nodes), [(k, "#") for k in H.nodes])))
                continue

            elif i != "#" and j != "#":
                tree_result = self.__tracetree.loc[i, j]
                forest_result = self.__traceforest.loc[i, j]

                if tree_result[0] == 0:
                    if tree_result[1][0] == "#" and tree_result[1][1] != "#":
                        H = nx.dfs_tree(self.__tree2,
                                        source=tree_result[1][1])
                        Hprime = nx.dfs_tree(
                            self.__tree2, source=tree_result[2][1])
                        H.remove_nodes_from(list(Hprime.nodes))
                        G = nx.compose(G, H)
                        G.add_edge(parent, tree_result[1][1])
                        G = nx.relabel_nodes(
                            G, dict(zip(list(H.nodes), [("#", k) for k in H.nodes])))

                    elif tree_result[1][0] != "#" and tree_result[1][1] == "#":
                        H = nx.dfs_tree(self.__tree1,
                                        source=tree_result[1][0])
                        Hprime = nx.dfs_tree(
                            self.__tree1, source=tree_result[2][0])
                        H.remove_nodes_from(list(Hprime.nodes))
                        G = nx.compose(G, H)
                        G.add_edge(parent, tree_result[1][0])
                        G = nx.relabel_nodes(
                            G, dict(zip(list(H.nodes), [(k, "#") for k in H.nodes])))

                    stack.append(
                        [tree_result[2][0], tree_result[2][1], tree_result[1]])

                elif tree_result[0] == 1:
                    G.add_node(tree_result[1])
                    G.add_edge(parent, tree_result[1])

                    foreststack = deque()
                    foreststack.append([i, j, (i, j)])

                    while len(foreststack) != 0:
                        i_f, j_f, p_f = foreststack.pop()

                        if i_f == "#" and j_f == "#":
                            continue

                        elif i_f == "#":
                            if j_f != "#":
                                for j_tmp in j_f:
                                    H = nx.dfs_tree(
                                        self.__tree2, source=j_tmp)
                                    G = nx.compose(G, H)
                                    G.add_edge(p_f, j_tmp)
                                    G = nx.relabel_nodes(
                                        G, dict(zip(list(H.nodes), [("#", k) for k in H.nodes])))

                        elif j_f == "#":
                            if i_f != "#":
                                for i_tmp in i_f:
                                    H = nx.dfs_tree(
                                        self.__tree1, source=i_tmp)
                                    G = nx.compose(G, H)
                                    G.add_edge(p_f, i_tmp)
                                    G = nx.relabel_nodes(
                                        G, dict(zip(list(H.nodes), [(k, "#") for k in H.nodes])))

                        elif i_f != "#" and j_f != "#":
                            i_f = "{}".format(i_f)
                            j_f = "{}".format(j_f)

                            forest_result = self.__traceforest.loc[i_f, j_f]

                            if forest_result[0] == 1:
                                stack.append(
                                    [forest_result[2][0], forest_result[2][1], p_f])
                                foreststack.append(
                                    [forest_result[1][0], forest_result[1][1], p_f])

                            elif forest_result[0] == 2:
                                foreststack.append(
                                    [forest_result[1][0][0], forest_result[1][0][1], p_f])
                                if forest_result[1][1][1] != "#":
                                    G.add_node(forest_result[2])
                                    G.add_edge(p_f, forest_result[2])
                                    foreststack.append(
                                        [forest_result[1][1][0], forest_result[1][1][1], forest_result[2]])

                                elif forest_result[1][1][1] == "#":
                                    H = nx.dfs_tree(
                                        self.__tree1, forest_result[1][1][0])
                                    G = nx.compose(G, H)
                                    G.add_edge(p_f, forest_result[1][1][0])
                                    G = nx.relabel_nodes(
                                        G, dict(zip(list(H.nodes), [(k, "#") for k in H.nodes])))

                            elif forest_result[0] == 3:
                                foreststack.append(
                                    [forest_result[1][0][0], forest_result[1][0][1], p_f])

                                if forest_result[1][1][0] != "#":
                                    G.add_node(forest_result[2])
                                    G.add_edge(p_f, forest_result[2])
                                    foreststack.append(
                                        [forest_result[1][1][0], forest_result[1][1][1], forest_result[2]])

                                elif forest_result[1][1][0] == "#":
                                    H = nx.dfs_tree(
                                        self.__tree2, forest_result[1][1][1])
                                    G = nx.compose(G, H)
                                    G.add_edge(p_f, forest_result[1][1][1])
                                    G = nx.relabel_nodes(
                                        G, dict(zip(list(H.nodes), [("#", k) for k in H.nodes])))

        G.remove_node("tempnode")
        alignmentcost = self.__treedistance.loc[self.__postorder1[-1],
                                                self.__postorder2[-1]]

        self.__alignmentcost = alignmentcost

        return G



import math
import os
import sys
import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc
import scanpy.external as sce
import anndata
from scipy import stats

def read_file(filename, transpose=False):
    adata = None
    if os.path.exists(filename):
        if os.path.isdir(filename):
            adata = sc.read_10x_mtx(filename)

        elif os.path.isfile(filename):
            name, filetype = os.path.splitext(filename)
            if filetype == ".txt":
                print()
                adata = sc.read_text(filename)

            if filetype == ".csv":
                adata = sc.read_csv(filename)

            if filetype == ".h5ad":
                adata = sc.read(filename)

        else:
            print(
                "ERROR: the format must be [H5AD|CSV|TXT] for file or 10x-MTX for directory.")
            sys.exit()

        if transpose:
            adata = adata.transpose()
    elif not os.path.exists(filename):
        sys.exit("ERROR: no such file or directory.")

    if not isinstance(adata.X, np.ndarray):
        X = adata.X.toarray()
        adata = anndata.AnnData(X, obs=adata.obs, var=adata.var)
    return adata

class Preprocessing:
    def __init__(self):
        pass

    def preprocessing_rawdata(
        self,
        adata,
        Min_Genes=200,
        Min_Cells=3,
        Min_Mean=0.0125,
        Max_Mean=3,
        Min_Disp=0.5,
        N_pcs=50,
        n_Top_genes=2000,
        K=10,
        magic_imputation=False,
        copy=False
    ):

        # adata = adata.copy() if copy else adata
        sc.pp.filter_cells(adata, min_genes=Min_Genes)
        sc.pp.filter_genes(adata, min_cells=Min_Cells)
        sc.pp.normalize_total(adata, exclude_highly_expressed=True)
        sc.pp.log1p(adata)

        if magic_imputation is True:
            #     set np.random.seed for magic to create same imputation
            np.random.seed(1)
            sce.pp.magic(
                adata,
                name_list='all_genes',
                knn=5
            )

        sc.pp.highly_variable_genes(
            adata,
            min_mean=Min_Mean,
            max_mean=Max_Mean,
            min_disp=Min_Disp,
            n_top_genes=n_Top_genes
        )
        adata.raw = adata
        # same as adata = adata[:,adata.var['highly_variable']] but inplace
        adata._inplace_subset_var(adata.var['highly_variable'])
        sc.tl.pca(
            adata,
            n_comps=N_pcs
        )
        sc.pp.neighbors(adata,
                        n_neighbors=K,
                        n_pcs=N_pcs
                        )
        sc.tl.diffmap(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata)
        sc.tl.paga(adata, groups='leiden')

        return adata if copy else None

    # get source node of tree
    # calcurating variance of gene expression level and
    # set node which has max variance as source node
    def calculate_root_node(
        self,
        adata,
        threshold,
        groupby="leiden"
    ):

        cluster_centroid_data = self.calculate_cluster_centroid(
            adata,
            dimension='raw',
            groupby=groupby
        )

        cluster_centroid = pd.DataFrame(
            cluster_centroid_data, index=adata.obs[groupby].cat.categories)

        cluster_centroid = cluster_centroid.loc[:,
                                                (cluster_centroid.sum(axis=0) != 0)]
        cor, pval = stats.spearmanr(cluster_centroid)
        gene_adjacency_matrix = np.where(abs(cor) > threshold, True, False)
        scEnergy = []
        for i in range(len(cluster_centroid.index)):
            tmp = 0
            for j in range(len(cluster_centroid.columns)):
                if cluster_centroid.iloc[i, j] != 0:
                    tmp += -cluster_centroid.iloc[i, j]*math.log(
                        cluster_centroid.iloc[i, j]/cluster_centroid.iloc[i, :][gene_adjacency_matrix[j]].sum())
            scEnergy.append(tmp)
        root_node = cluster_centroid.index[scEnergy.index(max(scEnergy))]

        return root_node

    # cluster_centroid: pd.DataFrame
    # index is cluster name, columns is gene name, X is gene expression level
    # argument "dimension" is either pca or diffmap

    def calculate_cluster_centroid(
        self,
        adata,
        dimension="pca",
        groupby="leiden"
    ):

        if dimension == "pca":
            X_dimension = "X_pca"
        elif dimension == "diffmap":
            X_dimension = "X_diffmap"
        elif dimension == "raw":
            X_dimension = "raw"
        else:
            raise ValueError(
                "Argument 'dimension' must be 'pca' or 'diffmap'.")

        if dimension in ["pca", "diffmap"]:
            clustername = adata.obs[groupby].cat.categories
            cluster_centroid_data = np.empty(
                (0, adata.obsm[X_dimension].shape[1]))
            for i in clustername:
                a_cluster_data = pd.DataFrame(
                    adata[adata.obs[groupby] == "{}".format(i)].obsm[X_dimension])
                a_cluster_median = a_cluster_data.median(axis=0).values
                cluster_centroid_data = np.vstack(
                    (cluster_centroid_data, a_cluster_median))
        else:
            clustername = adata.obs[groupby].cat.categories
            cluster_centroid_data = np.empty(
                (0, adata.X.shape[1]))
            for i in clustername:
                a_cluster_data = pd.DataFrame(
                    adata[adata.obs[groupby] == "{}".format(i)].X)
                a_cluster_median = a_cluster_data.median(axis=0).values
                cluster_centroid_data = np.vstack(
                    (cluster_centroid_data, a_cluster_median))

        return cluster_centroid_data

    def _convert_to_tree(
        self,
        adata,
        root_node,
        cluster_centroid_data,
        method,
        groupby="leiden"
    ):
        if root_node in adata.obs[groupby].cat.categories:
            pass
        else:
            raise ValueError(
                "Root {} does not exist in the clusters. Choose the root again.".format(root_node))

        if method == "paga":
            Adjacency_matrix = np.array(
                adata.uns['paga']["connectivities_tree"].todense()
            )

            G = nx.from_numpy_matrix(Adjacency_matrix, create_using=nx.Graph())

            # get names of cluster from adata.uns['paga']['groups'] as string
            groups_key = adata.uns['paga']['groups']
            groups_names = adata.obs[groups_key].cat.categories

            # relabel nodes name as names set in adata.obs of adata.uns['paga']['groups']
            G = nx.relabel_nodes(G, dict(zip(G, groups_names)))

        elif method == "gauss":
            from scipy.spatial.distance import pdist, squareform
            ka = 6
            y = pdist(cluster_centroid_data, 'euclid')
            Y = squareform(y)
            sigma = np.sort(Y, axis=1)[:, ka]
            affinity_matrix = np.exp(np.square(Y/sigma) * -1)
            matrix = 1 - affinity_matrix
            G = nx.from_numpy_matrix(matrix, create_using=nx.Graph())
            G = nx.relabel_nodes(
                G, dict(zip(G, adata.obs[groupby].cat.categories)))

        elif method == "euclid":
            from scipy.spatial.distance import pdist, squareform
            y = pdist(cluster_centroid_data, 'euclid')
            Y = squareform(y)
            G = nx.from_numpy_matrix(Y, create_using=nx.Graph())
            G = nx.relabel_nodes(
                G, dict(zip(G, adata.obs[groupby].cat.categories)))

        T = nx.minimum_spanning_tree(G)
        sorted(T.edges(data=True))
        tree = nx.dfs_tree(T, root_node)

        return tree

    def _set_postorder_successors(
        self,
        tree,
        root_node
    ):

        postorder_tmp = list(nx.dfs_postorder_nodes(tree, source=root_node))
        postorder = np.array([str(i) for i in postorder_tmp], dtype=object)
        successors = dict(
                        zip(
                            map(str, list(tree.nodes)),
                            [np.array(list(tree.successors(i)), dtype=object) for i in tree]
                            )
                        )

        return postorder, successors

    # combining methods above
    def trajectory_tree(
        self,
        adata,
        root_node=None,
        tree=None,
        groupby="leiden",
        method="euclid",
        dimension="pca",
        copy=False
    ):

        # if adata.X is scipy.sparsematrix, adata.X is converted to numpy array.
        if not isinstance(adata.X, np.ndarray):
            adata_tmp = adata.copy()
            adata_tmp.X = adata.X.toarray()

        else:
            adata_tmp = adata.copy()
        # add numpy array:cluster centroids matrix to adata.uns["cluster_centroid"]
        # index of cluster_centorid is adata.obs["leiden"].cat.categories
        cluster_centroid_data = self.calculate_cluster_centroid(
                                adata_tmp,
                                dimension=dimension,
                                groupby=groupby
                            )
        adata.uns["cluster_centroid"] = cluster_centroid_data

        # add networkx.classes.digraph.DiGraph of trajectory to adata.uns["tree"]["tree"]
        # add postorder list to adata.uns["tree"]["postorder"]
        # add successors dictionary to adata.uns["tree"]["successors"]
        # add root node str to adata.uns["tree"]["root_node"]
        adata.uns["capital"] = {}
        adata.uns["capital"]["tree"] = {}

        tree_dict = adata.uns["capital"]["tree"]
        tree_dict["annotation"] = groupby

        if tree is None:
            if root_node is not None:
                tree_dict["root_node"] = root_node
            else:
                root_node = self.calculate_root_node(
                    adata_tmp, threshold=0.5, groupby=groupby)
                tree_dict["root_node"] = root_node
                print("{} is set as a root node".format(root_node))

            tree = self._convert_to_tree(
                    adata_tmp,
                    root_node,
                    cluster_centroid_data,
                    method,
                    groupby=groupby
                    )

            postorder, successors = self._set_postorder_successors(tree, root_node)

            tree_dict["tree"] = nx.convert_matrix.to_pandas_adjacency(
                tree, nodelist=tree.nodes())
            tree_dict["postorder"] = postorder
            tree_dict["successors"] = successors

        else:
            if type(tree) is nx.DiGraph:
                root_node_from_tree = list(nx.topological_sort(tree))[0]
                if root_node is None:
                    root_node = root_node_from_tree
                    postorder, successors = self._set_postorder_successors(tree, root_node)
                elif root_node is not None:
                    if root_node_from_tree == root_node:
                        postorder, successors = self._set_postorder_successors(tree, root_node)
                    else:
                        raise ValueError("root node of tree and passed argument 'root_node' did not match.")

            elif type(tree) is nx.Graph and root_node is not None:
                tree = nx.dfs_tree(tree, root_node)
                postorder, successors = self._set_postorder_successors(tree, root_node)
            else:
                raise ValueError("Argument 'tree' must be nx.DiGraph or nx.Graph with 'root_node' argument passed")

            tree_dict["root_node"] = root_node
            tree_dict["tree"] = nx.convert_matrix.to_pandas_adjacency(
                tree, nodelist=tree.nodes())
            tree_dict["postorder"] = postorder
            tree_dict["successors"] = successors

        return adata if copy else None



from setuptools import setup, find_packages
from os import path
import re

package_name = "capital"
root_dir = path.abspath(path.dirname(__file__))

with open(path.join(root_dir, "README.md"), "r") as fh:
    long_description = fh.read()


with open(path.join(root_dir, package_name, '__init__.py')) as f:
    init_text = f.read()
    version = re.search(
        r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    license = re.search(
        r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(
        r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(
        r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)


setup(
    name=package_name,
    packages=find_packages(),
    version=version,
    license=license,
    install_requires=[
        name.rstrip() for name in
        open(path.join(root_dir, 'requirements.txt')).readlines()
        ],
    author=author,
    author_email=author_email,
    url=url,
    description='Single-Cell Analysis, comparing pseudotime trajectories \
                with tree alignment',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Framework :: Jupyter',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)
